# -----------------------------------------------------------
# Copyright (c) 2018 Lenovo AI Platform. All Rights Reserved.
# -----------------------------------------------------------

'''
Example: 
For imagenet dataset, model resnet
# Infer a single img file
python class_inference.py --infer_filename='/home/xh/data/pics/dog1.jpg' --model_name='resnet_v2_50' --checkpoint_dir='/home/xh/data/resnet-50' --output_dir=/tmp
# Infer a img file list
python class_inference.py --input_dir='/home/xh/data/pics/' --model_name='resnet_v2_50' --checkpoint_dir='/home/xh/data/resnet-50' --output_dir=/tmp
'''

from __future__ import division

import math
import time
import os
import sys
from datetime import datetime
from os.path import join, isfile, splitext

import numpy as np
import tensorflow as tf
import tensorflow.python.platform
from tensorflow.python.eager import context
from tensorflow.python.platform import gfile
parentdpath = os.path.dirname(os.path.abspath(__file__))
path_list = [('/').join(parentdpath.split('/')[:-2]) + '/lib', ('/').join(parentdpath.split('/')[:-2]) + '/engine', parentdpath]
sys.path += path_list
from datasets.tf_record_converter import get_train_data

parentpath = os.path.dirname(os.path.abspath(__file__))[:-24]

path_list = [parentpath + 'lib',
             parentpath + 'engine',
             parentpath + 'inference',
             parentpath + 'applications']
sys.path += path_list
print sys.path

tf.app.flags.DEFINE_string('infer_filename', '',
                           """File where to read target inference image file """)
import input_reader_orig as input_reader
import inference_util
from nets import nets_factory
from preprocessing import decode_input, preprocessing_factory, image_preprocessing
from preprocessing import image_input_handler as handler

slim = tf.contrib.slim
FLAGS = tf.app.flags.FLAGS
tf.logging.set_verbosity(tf.logging.INFO)

LABELS_FILE = FLAGS.checkpoint_dir + '/labels.txt'


def pre_check(label_file=LABELS_FILE):
    '''pre check all flags to make sure following work can be executed.
    '''
    if not FLAGS.model_name:
      raise ValueError("Inference model name must be provided.")
    if not FLAGS.checkpoint_dir:
      raise ValueError("Checkpoint dir must be provided to restore model.")
    if not gfile.Exists(label_file):
      raise ValueError("label file: %s is not found." % label_file)
    if not FLAGS.infer_filename and not FLAGS.input_dir:
      raise ValueError("Target inference file must be provided.")
    if FLAGS.infer_filename and FLAGS.input_dir:
      raise ValueError("FLAG infer_filename and input_dir can't be provided both.")
    if not FLAGS.output_dir:
      raise ValueError("Output dir must be provided to restore predict results.")

def load(saver, sess, ckpt_path):
    '''Load trained weights.
    
    Args:
      saver: TensorFlow saver object.
      sess: TensorFlow session.
      ckpt_path: path to checkpoint file with parameters.
    '''
    saver.restore(sess, ckpt_path)
    print("Restored model parameters from {}".format(ckpt_path))
def get_image(_data, height, width, _fn):
    if FLAGS.data_format == 'raw_image' and FLAGS.task == 'custom':
      resize_side = int(height * 1.15)      
      image = image_preprocessing.preprocess_image(_data.uint8image, height, width, resize_side=resize_side)    
    elif FLAGS.data_format == 'tfrecord' and FLAGS.task == 'custom':
      raise Exception("ERROR custom data not support tfrecrd data")
      image = get_train_data(height, resize_side)[0]
    else:
      image = _fn(_data.uint8image, height, width)
    return image

def preprocess_file(filename, height, width):
    """Preprocess data from single image file 

    Args:
        filename: image file name (full path)
        height: expected file size
        width: expected file size

    Return:
        image with expected height/width using corresponding nets model.
    """

    class Record(object):
        pass

    # decode img
    #_data = decode_input.read_data_file(filename)
    _data = Record()
    _data.name = filename
    file_contents = tf.read_file(_data.name)
    record_bytes = tf.image.decode_image(file_contents, channels=3, name='image_decode')
    _data.uint8image = record_bytes

    _name = FLAGS.model_name
    _fn = preprocessing_factory.get_preprocessing(_name, is_training=False)
    image = get_image(_data, height, width, _fn)
    return tf.reshape(image, [1,height,width,3])

def preprocess_filelist(filenamelist, height, width, num_batch):
    """Preprocess data from image file list

    Args:
        filename_list: image file name list (full path)
        height: expected file size
        width: expected file size
        num_batch: number of files in filename_list

    Return:
        images with expected height/width using corresponding nets model.
    """

    class Record(object):
        pass

    #_data_list = decode_input.read_data_file_list(filenamelist, num_batch)
    _data_list = []
    for i in range(num_batch):
        filename = filenamelist[i]
        result = Record()
        result.name = filename
        file_contents = tf.read_file(result.name)
        record_bytes = tf.image.decode_image(file_contents, channels=3, name='image_decode')
        result.uint8image = record_bytes
        _data_list.append(result)

    _name = FLAGS.model_name
    _reshaped_img_list = []
    _fn = preprocessing_factory.get_preprocessing(_name, is_training=False)
    for _data in _data_list:    
      image = get_image(_data, height, width, _fn)
      image = tf.reshape(image, [height,width,3])
      _reshaped_img_list.append(image)
    return tf.stack(_reshaped_img_list)
   
def get_available_device():
    num_gpu = context.num_gpus()
    return '/device:GPU:0' if num_gpu else '/device:CPU:0'

def inference():
    """Inference from a single image file or an image file list,
       this is only supported for models under nets/.
       Only imagenet and customized datasets supported currently.

    Return:
        Predict target labels from previous train welled checkpoints
        with the full label files.
    """
    with tf.Graph().as_default():

      # Get target label doc for matching probability
      labels_dict = {}
      use_csv = False
      try:
        with open(LABELS_FILE) as f:
          for line in f:
            key, value = line.rstrip('\n').split(':')
            labels_dict[key] = value
      except IOError:
        raise ValueError("Can't open file:", LABELS_FILE)

      infer_device = get_available_device()
      with tf.device(infer_device):
        # Get model 
        num_classes = len(labels_dict)
        network_fn = nets_factory.get_network_fn(FLAGS.model_name,
                                                 num_classes,
                                                 is_training=False)
        image_size= network_fn.default_image_size

        batch_size = FLAGS.batch_size
        num_batch = 0
        file_list, image_list = [], []
        if FLAGS.infer_filename:
          file_list = [FLAGS.infer_filename]
          image_list.append(preprocess_file(FLAGS.infer_filename, image_size, image_size))
        elif FLAGS.input_dir:
          files = os.walk(FLAGS.input_dir)
          for _path,_dir_list,_file_list in files:  
              for _file_name in _file_list:  
                if 'csv' in _file_name:
                    use_csv = True
                    input_files = os.path.join(_path, _file_name)
          if use_csv:
              with open(input_files, 'r') as f:
                  files_path = f.readlines()
              for file_path in files_path:
                  if file_path.split()[0].startswith('/'):
                      file_path = file_path.split()[0]
                  else:
                      file_path = '/'.join(input_files.split('/')[:-1]) + '/' + file_path.split()[0]
                  file_list.append(file_path)
                  batch_size -= 1
                  num_batch += 1
                  if batch_size == 0:
                      batch_size = FLAGS.batch_size
                      image_list.append(preprocess_filelist(file_list[num_batch-batch_size:num_batch], image_size, image_size, batch_size))
                      print('Image processing is {} % complete'.format(round((len(file_list)/len(files_path)),3)*100))
                  if num_batch == len(files_path)-1:
                      batch_size = FLAGS.batch_size - batch_size
                      image_list.append(preprocess_filelist(file_list[num_batch-batch_size:num_batch], image_size, image_size, batch_size))
          else:
              file_list = inference_util.get_img_files(FLAGS.input_dir)
              num_batch = len(file_list)
              image = preprocess_filelist(file_list, image_size, image_size, num_batch)


        # Predictions
        if use_csv:
            image_place = tf.placeholder(tf.float32, shape=(None, image_size, image_size, 3), name='image_place')
            logits, _ = network_fn(image_place)
        else:
            logits, _ = network_fn(image)

        # Calculate probabilities with softmax activation
        
        prob = tf.nn.softmax(logits)

        # Create session 
        gpu_options = tf.GPUOptions(per_process_gpu_memory_fraction=0.7)
        sess = tf.Session(config=tf.ConfigProto(gpu_options=gpu_options,
                                                allow_soft_placement=True,
                                                log_device_placement=True))
 
        # Restore variables from checkpoint file into graph
        saver = tf.train.Saver()
        ckpt = tf.train.latest_checkpoint(FLAGS.checkpoint_dir)
        load(saver, sess, ckpt)

        # Code below are used for actual op compute
        if use_csv:
            predict_picture = 0
            for image in image_list:
                _logits_val, _prob_val = sess.run([logits,prob],feed_dict={image_place:sess.run(image)})
                predict_picture += 1
                print ('Picture predict {}% complete'.format(round((predict_picture/len(image_list)),3)*100))
                try:
                    logits_val=np.vstack((logits_val, _logits_val))
                    prob_val=np.vstack((prob_val, _prob_val))
                except:
                    logits_val = _logits_val
                    prob_val = _prob_val

            predict_res = {}
            if isinstance(prob_val, np.ndarray):
              probs = []
              labels = []
              index = np.argmax(prob_val, axis=1)
              for i, j in enumerate(index):
                probs.append(prob_val[i][j])
                labels.append(labels_dict[str(j)])
              right, error = 0, 0
              every_label_acc = {}
              for f, label in zip(file_list, labels):
                predict_res[f] = label
                print("File:[%s]  label:[%s] \n" % (f, label))
                real_label = f.split('/')[-2]
                if f.split('/')[-2] == label:
                    if real_label in every_label_acc.keys():
                        every_label_acc[real_label]=map(lambda x:x+1,every_label_acc[real_label])
                    else:
                        every_label_acc[real_label]=[1, 1]
                    right += 1
                else:
                    if real_label in every_label_acc.keys():
                        every_label_acc[real_label][1] += 1
                    else:
                        every_label_acc[real_label]=[0, 1]
                    error += 1
              acc_log = ''
              for label, result in every_label_acc.items():
                  acc_log += 'accuracy:' + '{} accuracy {} \n'.format(label, result[0]/result[1])
              acc_log += 'all_right: {}, all_error: {} \n'.format(right, error)
              acc_log += 'Average accuracy: {} \n'.format(right/(right+error))
              print(acc_log)
              inference_util.output_to_file(
                  predict_res, FLAGS.output_dir + '/predict_result.txt')
              inference_util.output_to_file(
                  acc_log, FLAGS.output_dir + '/accuracy.txt')
            else:
              raise ValueError("Unknown format probability result. ")
        else:
            logits_val, prob_val = sess.run([logits, prob])

            predict_res = {}
            if isinstance(prob_val, np.ndarray):
                probs = []
                labels = []
                index = np.argmax(prob_val, axis=1)
                for i, j in enumerate(index):
                    probs.append(prob_val[i][j])
                    labels.append(labels_dict[str(j)])
                # print("Probabilities are: %s" % probs)
                for f, label in zip(file_list, labels):
                    predict_res[f] = label
                    print("File:[%s]  label:[%s] \n" % (f, label))

                inference_util.output_to_file(
                    predict_res, FLAGS.output_dir + '/predict_result.txt')
            else:
                raise ValueError("Unknown format probability result. ")

def main(argv=None):  # pylint: disable=unused-argument
    pre_check()
    inference()
if __name__ == '__main__':
    tf.app.run()

